/*
******************************************************************************

   LibBlt: extension initialization.

   Copyright (C) 2000 Christoph Egger   [Christoph_Egger@t-online.de]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <string.h>

#include "config.h"
#include <ggi/internal/blt.h>
#include <ggi/internal/blt_debug.h>

/* Static variables 
 */
static int 	 _bltLibIsUp       = 0;
static void	*_bltConfigHandle  = NULL;
static char      _bltconfstub[512] = BLTCONFFILE;
static char     *_bltconffile      = _bltconfstub+BLTTAGLEN;

static ggiGA_resource_list reqdummy = NULL;

int _bltDebugSync = 0;
uint32 _bltDebugState = 0;


/* Extension ID. Defaulting to -1 should make segfault on abuse more likely...
 * This is exported so that sublibs can use it.
 */
EXPORTVAR ggi_extid _bltID = -1;


#define SUBLIB_PREFIX	"BLTdl_"

/* Dummy function which returns -1
   We use this to reset the function pointers */
static int dummyfunc(void)
{
	printf("LibBLT: dummyfunc() of LibBLT called\n");

	return GALLOC_EUNAVAILABLE;
}	/* dummyfunc */


/* Reset all function pointers to dummyfunc */
static void clearfuncs(bltpriv *priv)
{
	priv->tie =   (void *)dummyfunc;
	priv->untie = (void *)dummyfunc;
}	/* clearfuncs */


/* This function is called by LibGGI when the API-list changes */
static int changed(ggi_visual_t vis, int whatchanged)
{
	printf("LibBLT: changed(%p, 0x%x) of LibBLT called\n",
	       vis, whatchanged);

	switch (whatchanged) {
	case GGI_CHG_APILIST: {
		int i;
		const char *fname;
		ggi_dlhandle *lib;
		char api[GGI_MAX_APILEN], args[GGI_MAX_APILEN];

		clearfuncs(LIBGGI_BLTEXT(vis));

		for(i=0; ggiGetAPI(vis, i, api, args) == 0; i++) {
			strcat(api,"-blt");
			fname = ggMatchConfig(_bltConfigHandle, api, NULL);
			if (fname == NULL) {
				/* No special implementation for this sublib */
				continue;
			}
			lib = ggiExtensionLoadDL(vis, fname, args, NULL,
						 SUBLIB_PREFIX);
		}	/* for */
	}
	break;
	}	/* switch */
                                
	return 0;
}	/* changed */


/* Initialize the extension
 */
int ggiBltInit(void)
{
	int err;

	_bltLibIsUp++;
	if (_bltLibIsUp > 1) return 0; /* Initialize only at first call. */

	err = ggLoadConfig(_bltconffile, &_bltConfigHandle);
	if (err != GGI_OK) {
		fprintf(stderr,"LibBLT: couldn't open %s\n",
			_bltconffile);
		_bltLibIsUp--;
		return err;
	}	/* if */

	_bltID = ggiExtensionRegister("BLT", sizeof(bltpriv), changed);

	if (_bltID < 0) {
		fprintf(stderr,
			"LibBLT: failed to register as extension.\n");
		_bltLibIsUp--;
		ggFreeConfig(_bltConfigHandle);
		return _bltID;
	}	/* if */

	ggiGAInit();

	return 0;
}	/* ggiBltInit */


/* Deinitialize the extension
 */
int ggiBltExit(void)
{
	int rc;
    
	if (!_bltLibIsUp) return -1;

	if (_bltLibIsUp > 1) {
		/* Exit only at last call */
		_bltLibIsUp--;
		return 0;
	}
	
	rc = ggiExtensionUnregister(_bltID);
	ggFreeConfig(_bltConfigHandle);

	_bltLibIsUp = 0;

	ggiGAExit();

	return rc;
}	/* ggiBltExit */


/* Attach the extension to a visual
 */
int ggiBltAttach(ggi_visual_t vis, ggiGA_resource_list *reqlist)
{
	int rc;

	printf("LibBLT: ggiBltAttach(%p) called\n", vis);

	rc = ggiExtensionAttach(vis, _bltID);
	if (rc < 0) {
		goto err0;
	}	/* if */

	if (rc == 0) {
		rc = ggiGAAttach(vis);
		if (rc < 0) {
			goto err1;
		}	/* if */
		rc = 0;

		/* We are actually creating the primary instance. */
		memset(LIBGGI_BLTEXT(vis), 0, sizeof(bltpriv));
		LIBGGI_BLTEXT(vis)->priv = NULL;
		if (reqlist == NULL) {
			LIBGGI_BLTEXT(vis)->reqlist = &reqdummy;
		} else {
			LIBGGI_BLTEXT(vis)->reqlist = reqlist;
		}	/* if */

		/* Now fake an "API change" so the right libs get loaded */
		changed(vis, GGI_CHG_APILIST);
	}	/* if */

	return rc;

err1:
	ggiExtensionDetach(vis, _bltID);
err0:
	return rc;
}	/* ggiBltAttach */


/* Detach the extension from a visual
 */
int ggiBltDetach(ggi_visual_t vis)
{
	int rc;
	void *priv;
	priv = LIBGGI_BLTEXT(vis)->priv;

	printf("LibBlt: ggiBltDetach(%p) called\n", vis);

	rc = ggiExtensionDetach(vis, _bltID);
	if (rc != 0) return rc;	

	if (priv != NULL) {
		free(priv);
	}	/* if */

	ggiGADetach(vis);

	return rc;
}	/* ggiBltDetach */

/* This should be included by exactly _one_ file in each extension library. */
#include <ggi/internal/ggilibinit.h>
